""" Blueprint for querying job statuses

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

This module defines a Flask Blueprint for querying the current status of a specific job or all logged jobs.

"""


from flask import Blueprint, request, jsonify
from nova_server.utils.job_utils import JOBS, get_all_jobs, JobStatus
from nova_server.utils.job_utils import get_job_id_from_request_form

status = Blueprint("status", __name__)


@status.route("/job_status", methods=["POST"])
def job_status():
    """
    Query the status of a specific job.

    This route allows querying the status of a job by providing the job's unique identifier in the request.

    Returns:
        dict: A JSON response containing the status of the requested job.

    Example:
        >>> POST /job_status
        >>> {"job_id": "12345"}
        {"status": "RUNNING"}
    """
    if request.method == "POST":
        request_form = request.form.to_dict()
        status_key = get_job_id_from_request_form(request_form)

        if status_key in JOBS.keys():
            status = JOBS[status_key].status
            return jsonify({"status": status.value})
        else:
            return jsonify({"status": JobStatus.WAITING.value})


@status.route("/job_status_all", methods=["GET"])
def job_status_all():
    """
     Query the status of all jobs.

     This route allows querying the status of a job by providing the job's unique identifier in the request.

     Returns:
         dict: A JSON response containing the status of the requested job.

     Example:
         >>> POST /job_status
         >>> {"job_id": "12345"}
         [{
            start_time = '09/13/2023, 16:51:27'
            end_time = None
            progress = 'Predicting data'
            status = JobStatus.RUNNING
            job_key = 12345
            interactive_url = interactive_url
            log_path = /logs/12345.log
            details = {'user' : 'my_user', 'dataset' : 'test' ... }
         }, ...]
     """
    return jsonify(get_all_jobs())



"""Backend for using virtual environments as a backend for nova server

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    18.8.2023

"""

import os
import shutil
import subprocess

import sys
from threading import Thread
from subprocess import Popen, PIPE
from pathlib import Path
from nova_server.utils import venv_utils as vu
from dotenv import load_dotenv
from logging import Logger


class VenvHandler:
    """
    Handles the creation and management of a virtual environment and running scripts within it.

    Args:
        module_dir (Path, optional): The path to the nova server module directory for the environtment.
        logger (Logger, optional): The logger instance for logging.
        log_verbose (bool, optional): If True, log verbose output.

    Attributes:
        venv_dir (Path): The path to the virtual environment.
        log_verbose (bool): If True, log verbose output.
        module_dir (Path): The path to the module directory.
        logger (Logger): The logger instance.

    Example:
        >>> import logging
        >>> load_dotenv("../.env")
        >>> log_dir = Path(os.getenv('NOVA_LOG_DIR', '.'))
        >>> log_file = log_dir / 'test.log'
        >>> logging.basicConfig(filename=log_file, encoding='utf-8', level=logging.DEBUG)
        >>> logger = logging.getLogger('test_logger')
        >>> module_path = Path(os.getenv("NOVA_CML_DIR")) / "test"
        >>> venv_handler = VenvHandler(module_path, logger=logger, log_verbose=True)
        >>> venv_handler.run_python_script_from_file(
        ...     module_path / "test.py",
        ...     script_args=["pos_1", "pos_2"],
        ...     script_kwargs={"-k1": "k1", "--keyword_three": "k3"},
        ... )
    """

    def _reader(self, stream, context=None):
        """
        Reads and logs output from a stream.

        Args:
            stream: The stream to read.
            context (str, optional): The context of the stream (stdout or stderr).
        """
        while True:
            s = stream.readline()
            if not s:
                break
            if self.logger is None:
                if not self.log_verbose:
                    sys.stderr.write(".")
                else:
                    sys.stderr.write(s)
            else:
                if context == "stderr":
                    self.logger.error(s)
                else:
                    self.logger.info(s)
            sys.stderr.flush()
        stream.close()

    def _run_cmd(self, cmd: str, wait: bool = True) -> int:
        """
        Executes a command in a subprocess and logs the output.

        Args:
           cmd (str): The command to execute.
           wait (bool, optional): If True, wait for the command to complete.

        Returns:
            int: Return code of the executed command
        """
        p = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True, universal_newlines=True)
        t1 = Thread(target=self._reader, args=(p.stdout, "stdout"))
        t1.start()
        t2 = Thread(target=self._reader, args=(p.stderr, "stderr"))
        t2.start()
        if wait:
            p.wait()
            t1.join()
            t2.join()

        return p.returncode

    def _get_or_create_venv(self):
        """
        Gets or creates a virtual environment and returns its path.

        Returns:
            Path: The path to the virtual environment.
        """
        venv_dir = vu.venv_dir_from_mod(self.module_dir)
        if not venv_dir.is_dir():
            try:
                run_cmd = f"{sys.executable} -m venv {venv_dir}"
                self._run_cmd(run_cmd)
            except Exception as e:
                shutil.rmtree(venv_dir)
                raise e
        return venv_dir

    def _upgrade_pip(self):
        """
        Upgrades the `pip` package within the virtual environment.
        """
        run_cmd = vu.get_module_run_cmd(
            self.venv_dir, "pip", args=["install"], kwargs={"--upgrade": "pip"}
        )
        self._run_cmd(run_cmd)

    def _install_requirements(self):
        """
        Installs requirements from a `requirements.txt` file within the virtual environment.
        """
        req_txt = self.module_dir / "requirements.txt"
        if not req_txt.is_file():
            return
        else:
            # pip
            self._upgrade_pip()
            # TODO: replace with custom import utils for better parsing
            # requirements.txt
            run_cmd = vu.get_module_run_cmd(
                self.venv_dir,
                "pip",
                args=["install"],
                kwargs={"-r": str(req_txt.resolve())},
            )
            self._run_cmd(run_cmd)

    def __init__(
        self, module_dir: Path = None, logger: Logger = None, log_verbose: bool = False
    ):
        """
        Initializes the VenvHandler instance.

        Args:
            module_dir (Path, optional): The path to the module directory.
            logger (Logger, optional): The logger instance for logging.
            log_verbose (bool, optional): If True, log verbose output.
        """
        self.venv_dir = None
        self.log_verbose = log_verbose
        self.module_dir = module_dir
        self.logger = logger if logger is not None else Logger(__name__)
        if module_dir is not None:
            self.init_venv()

    def init_venv(self):
        """
        Initializes the virtual environment and installs requirements.
        """
        self.venv_dir = self._get_or_create_venv()
        self._install_requirements()

    def run_python_script_from_file(
        self, script_fp: Path, script_args: list = None, script_kwargs: dict = None
    ):
        """
        Runs a Python script within the virtual environment.

        Args:
            script_fp (Path): The path to the script to run.
            script_args (list, optional): List of positional arguments to pass to the script.
            script_kwargs (dict, optional): Dictionary of keyword arguments to pass to the script.

        Raises:
            ValueError: If the virtual environment has not been initialized. Call `init_venv()` first.
            subprocess.CalledProcessError: If the executed command exits with a return value other than 0
        """
        if self.venv_dir is None:
            raise ValueError(
                "Virtual environment has not been initialized. Call <init_venv()> first."
            )
        run_cmd = vu.get_python_script_run_cmd(
            self.venv_dir, script_fp, script_args, script_kwargs
        )
        return_code = self._run_cmd(run_cmd)
        if not return_code == 0:
            raise subprocess.CalledProcessError(returncode=return_code, cmd=run_cmd)

    def run_shell_script(self, script: str, script_args: list = None, script_kwargs: dict = None):
        """
        Runs a command in the respective os shell with an activated virtual environment


        Args:
            script (Path): The path to the script to run.
            script_args (list, optional): List of positional arguments to pass to the script.
            script_kwargs (dict, optional): Dictionary of keyword arguments to pass to the script.

        Raises:
            ValueError: If the virtual environment has not been initialized. Call `init_venv()` first.
            subprocess.CalledProcessError: If the executed command exits with a return value other than 0
        """
        if self.venv_dir is None:
            raise ValueError(
                "Virtual environment has not been initialized. Call <init_venv()> first."
            )
        run_cmd = vu.get_shell_script_run_cmd(
            self.venv_dir, script, script_args, script_kwargs
        )
        return_code = self._run_cmd(run_cmd)
        if not return_code == 0:
            raise subprocess.CalledProcessError(returncode=return_code, cmd=run_cmd)


if __name__ == "__main__":
    import logging

    load_dotenv("../.env")

    log_dir = Path(os.getenv("NOVA_LOG_DIR", "."))
    log_file = log_dir / "test.log"
    logging.basicConfig(filename=log_file, encoding="utf-8", level=logging.DEBUG)
    logger = logging.getLogger("test_logger")

    module_path = Path(os.getenv("NOVA_CML_DIR")) / "test"
    venv_handler = VenvHandler(module_path, logger=logger, log_verbose=True)
    venv_handler.run_python_script_from_file(
        module_path / "test.py",
        script_args=["pos_1", "pos_2"],
        script_kwargs={"-k1": "k1", "--keyword_three": "k3"},
    )



"""Utility modules for NOVA-Server Jobs

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    13.09.2023

"""

import copy
import random
from datetime import datetime
from enum import Enum
from nova_server.utils.thread_utils import status_thread_wrapper
from . import log_utils
from ..data import ANIMALS, COLORS


def get_job_id_from_request_form(request_form):
    """
    Returns the logging key from the provided request form
    Args:
        request_form (dict): Request form from Nova

    Returns:

    """
    key = request_form.get("jobID", None)

    if key is None:
        key = f"{get_random_name()}"

    return key


def get_random_name(
    combo=(COLORS, ANIMALS), separator: str = " ", style: str = "capital"
):
    if not combo:
        raise Exception("combo cannot be empty")

    random_name = []
    for word_list in combo:
        part_name = random.choice(word_list)
        if style == "capital":
            part_name = part_name.capitalize()
        if style == "lowercase":
            part_name = part_name.lower()
        if style == "uppercase":
            part_name = part_name.upper()
        random_name.append(part_name)
    return separator.join(random_name)


JOBS = {}


class JobStatus(Enum):
    WAITING = 0
    RUNNING = 1
    FINISHED = 2
    ERROR = 3


class Job:
    def __init__(self, job_key, interactive_url=None, log_path=None, details=None):
        self.start_time = None
        self.end_time = None
        self.progress = None
        self.status = JobStatus.WAITING
        self.job_key = job_key
        self.interactive_url = interactive_url
        self.log_path = log_path
        self.details = details

    def serializable(self):
        s = copy.deepcopy(vars(self))
        for key in s.keys():
            s[key] = str(s[key])
        return s


@status_thread_wrapper
def add_new_job(job_key, interactive_url=None, request_form=None):
    log_path = log_utils.get_log_path_for_thread(job_key)
    job_details = log_utils.get_log_conform_request(request_form)
    job = Job(job_key, interactive_url, log_path, details=job_details)
    JOBS[job_key] = job

    return True


@status_thread_wrapper
def remove_job(job_key):
    try:
        del JOBS[job_key]
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")


@status_thread_wrapper
def update_status(job_key, status: JobStatus):
    try:
        JOBS[job_key].status = status

        if status == status.RUNNING:
            JOBS[job_key].start_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        if status == status.FINISHED or status == status.ERROR:
            JOBS[job_key].end_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")


@status_thread_wrapper
def update_progress(job_key, progress: str):
    try:
        JOBS[job_key].progress = progress
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")


@status_thread_wrapper
def set_log_path(job_key, log_path):
    try:
        JOBS[job_key].log_path = log_path
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")


@status_thread_wrapper
def get_log_path(job_key):
    try:
        return JOBS[str(job_key)].log_path
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")


@status_thread_wrapper
def get_all_jobs():
    return [job.serializable() for job in JOBS.values()]



""" Blueprint for retrieving a job's log file

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

This module defines a Flask Blueprint for retrieving a job's log file.

"""

from flask import Blueprint, request, jsonify
from nova_server.utils import log_utils
from nova_server.utils.job_utils import get_job_id_from_request_form


log = Blueprint("log", __name__)


@log.route("/log", methods=["POST"])
def log_thread():
    """
    Retrieve the log file for a specific job.

    This route allows retrieving the log file for a job by providing the job's unique identifier in the request.

    Returns:
        dict: A JSON response containing the log file content.

    Example:
        >>> POST /log
        >>> {"job_id": "12345"}
        {"message": "Log file content here..."}
    """
    if request.method == "POST":
        request_form = request.form.to_dict()
        log_key = get_job_id_from_request_form(request_form)
        if log_key in log_utils.LOGS:
            logger = log_utils.LOGS[log_key]
            path = logger.handlers[0].baseFilename
            with open(path) as f:
                f = f.readlines()
            output = ""
            for line in f:
                output += line
            return jsonify({"message": output})
        else:
            return jsonify({"message": "No log for the given parameters found."})



""" Blueprint for querying server capabilities

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    14.09.2023

This module defines a Flask Blueprint for querying the server for its capabilities.
Returns a list of all available server-module trainer and chains in json format.

"""
import os
import json
from flask import Blueprint, request
from nova_server.utils import env
from nova_utils.utils.json_utils import ChainEncoder, TrainerEncoder
from nova_utils.utils.ssi_xml_utils import Chain, Trainer
from pathlib import Path
from glob import glob

cml_info = Blueprint("cml_info", __name__)


@cml_info.route("/cml_info", methods=["POST"])
def info():
    """
    Query the server to return information about all

    This route allows querying the server for all available nova-server modules

    Returns:
        dict: A JSON response containing the status of the requested job.

    Example:

    """
    if request.method == "POST":
        cml_dir = os.getenv(env.NOVA_SERVER_CML_DIR)

        if not cml_dir:
            return None

        cml_dir = str(Path(cml_dir).resolve())

        trainer_files = glob(cml_dir + '/**/*.trainer', recursive = True)
        chain_files = glob(cml_dir + '/**/*.chain', recursive = True)

        trainer_ok = {}
        trainer_faulty = {}
        chains_ok = {}
        chains_faulty = {}

        for tf in trainer_files:
            t = Trainer()
            try:
                t.load_from_file(tf)
                trainer_ok[tf] = json.dumps(t, cls=TrainerEncoder)
            except Exception as e:
                trainer_faulty[tf] = str(e)

        for cf in chain_files:
            c = Chain()
            try:
                c.load_from_file(cf)
                chains_ok[cf] = json.dumps(c, cls=ChainEncoder)
            except Exception as e:
                chains_faulty[cf] = str(e)

        return {
            'chains_ok' : chains_ok,
            'chains_faulty' : chains_faulty,
            'trainer_ok' : trainer_ok,
            'trainer_faulty' : trainer_faulty
        }






""" Class definition for th execution handler. The execution handler combines uses data from the request form to execute
the correct command using the specified backends.

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023
"""

import os
import re
from logging import Logger
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from pathlib import PureWindowsPath
from nova_server.utils import env


class Action(Enum):
    PREDICT = "nu-predict"
    EXTRACT = "nu-extract"
    TRAIN = "nu-train"


class Backend(Enum):
    VENV = 0


class ExecutionHandler(ABC):
    @property
    def script_arguments(self):
        return self._script_arguments

    @script_arguments.setter
    def script_arguments(self, value):
        # convert dictionary keys from camelCase to snake_case
        self._script_arguments = {
            "--" + re.sub(r"(?<!^)(?=[A-Z])", "_", k).lower(): v
            for k, v in value.items()
        }

    def __init__(
        self, request_form: dict, backend: Backend = Backend.VENV, logger: Logger = None
    ):
        self.backend = backend
        self.script_arguments = request_form
        self.logger = logger

    def _nova_server_env_to_arg(self):
        env_vars = [env.NOVA_SERVER_CML_DIR, env.NOVA_SERVER_DATA_DIR, env.NOVA_SERVER_LOG_DIR, env.NOVA_SERVER_CACHE_DIR, env.NOVA_SERVER_TMP_DIR]
        arg_vars = {}
        prefix = 'NOVA_SERVER_'
        for var in env_vars:
            k = "--" + var[len(prefix):].lower()
            v = os.getenv(var)
            arg_vars[k] = v
        return arg_vars

    def run(self):

        # run with selected backend
        if self.backend == Backend.VENV:
            from nova_server.backend import virtual_environment as backend

            # setup virtual environment
            cml_dir = os.getenv(env.NOVA_SERVER_CML_DIR)
            if cml_dir is None:
                raise ValueError(f"NOVA_CML_DIR not et")

            module_dir = Path(cml_dir) / self.module_name
            if not module_dir.is_dir():
                raise NotADirectoryError(
                    f"NOVA_CML_DIR {module_dir} is not a valid directory"
                )
            backend_handler = backend.VenvHandler(
                module_dir, logger=self.logger, log_verbose=True
            )

            # add dotenv variables to arguments for script
            self._script_arguments |= self._nova_server_env_to_arg()

            backend_handler.run_shell_script(
                self.run_script,
                script_kwargs=self._script_arguments,
            )

        else:
            raise ValueError(f"Unknown backend {self.backend}")

    @property
    @abstractmethod
    def run_script(self):
        pass

    @property
    @abstractmethod
    def module_name(self):
        pass


class NovaPredictHandler(ExecutionHandler):
    @property
    def module_name(self):
        tfp = self.script_arguments.get("--trainer_file_path")
        if tfp is None:
            raise ValueError("trainerFilePath not specified in request.")
        else:
            return PureWindowsPath(tfp).parent

    @property
    def run_script(self):
        return self.action.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.action = Action.PREDICT

class NovaExtractHandler(ExecutionHandler):
    @property
    def module_name(self):
        cfp = self.script_arguments.get("--chain_file_path")
        if cfp is None:
            raise ValueError("chainFilePath not specified in request.")
        else:
            return PureWindowsPath(cfp).parent

    @property
    def run_script(self):
        return self.action.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.action = Action.EXTRACT

class NovaTrainHandler(ExecutionHandler):
    @property
    def module_name(self):
        tfp = self.script_arguments.get("--trainer_file_path")
        if tfp is None:
            raise ValueError("trainerFilePath not specified in request.")
        else:
            return PureWindowsPath(tfp).parent

    @property
    def run_script(self):
        return self.action.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.action = Action.TRAIN


"""This module lists all required environment variables used by NOVA-Server

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    13.09.2023

"""

NOVA_SERVER_HOST = "NOVA_SERVER_HOST"
NOVA_SERVER_PORT = "NOVA_SERVER_PORT"
NOVA_SERVER_CML_DIR = "NOVA_SERVER_CML_DIR"
NOVA_SERVER_DATA_DIR = "NOVA_SERVER_DATA_DIR"
NOVA_SERVER_CACHE_DIR = "NOVA_SERVER_CACHE_DIR"
NOVA_SERVER_TMP_DIR = "NOVA_SERVER_TMP_DIR"
NOVA_SERVER_LOG_DIR = "NOVA_SERVER_LOG_DIR"



"""Standalone script for data predictions

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

This script performs predictions for annotations in the NOVA database using a provided nova-server module for inference and saves the results to the NOVA-DB.

Args:
    --env (str, optional): Path to the environment file to read configuration from. Default: ''
    --host (str, optional): The host IP address for the server. Default: '0.0.0.0'
    --port (str, optional): The port on which the server will listen. Default: '8080'
    --cml_dir (str, optional): CML folder to read the training scripts from, same as in Nova. Default: 'cml'
    --data_dir (str, optional): Data folder to read the training scripts from, same as in Nova. Default: 'data'
    --cache_dir (str, optional): Cache folder where all large files (e.g., model weights) are cached. Default: 'cache'
    --tmp_dir (str, optional): Folder for temporary data storage. Default: 'tmp'
    --log_dir (str, optional): Folder for temporary data storage. Default: 'log'

Returns:
    None

Example:
    >>> app.py --host 0.0.0.0 --port 53771 --cml_dir "/path/to/my/cml" --data_dir "/path/to/my/data" --cache_dir "/path/to/my/cache" --tmp_dir "/path/to/my/tmp"
"""
import dotenv
from flask import Flask
from nova_server.route.train import train
from nova_server.route.extract import extract
from nova_server.route.status import status
from nova_server.route.log import log
from nova_server.route.ui import ui
from nova_server.route.cml_info import cml_info
from nova_server.route.cancel import cancel
from nova_server.route.predict import predict
import argparse
import os
from pathlib import Path
from waitress import serve
from nova_server.utils import env

print("Starting nova-backend server...")
app = Flask(__name__, template_folder="./templates")
app.register_blueprint(train)
app.register_blueprint(predict)
app.register_blueprint(extract)
app.register_blueprint(log)
app.register_blueprint(status)
app.register_blueprint(ui)
app.register_blueprint(cancel)
app.register_blueprint(cml_info)

parser = argparse.ArgumentParser(
    description="Commandline arguments to configure the nova backend server"
)
parser.add_argument("--env", type=str, default='', help="Path to the environment file to read config from")
parser.add_argument("--host", type=str, default="0.0.0.0", help="The host ip address")
parser.add_argument(
    "--port", type=str, default='8080', help="The port the server listens on"
)
parser.add_argument(
    "--cml_dir",
    type=str,
    default="cml",
    help="Cml folder to read the training scripts from. Same as in Nova.",
)
parser.add_argument(
    "--data_dir",
    type=str,
    default="data",
    help="Data folder to read the training scripts from. Same as in Nova.",
)
parser.add_argument(
    "--cache_dir",
    type=str,
    default="cache",
    help="Cache folder where all large files (e.g. model weights) are cached.",
)

parser.add_argument(
    "--tmp_dir",
    type=str,
    default="tmp",
    help="Folder for temporary data storage.",
)
parser.add_argument(
    "--log_dir",
    type=str,
    default="log",
    help="Folder for temporary data storage.",
)

def _run():

    # TODO: support multiple (data) directories
    args = parser.parse_args()
    default_args = parser.parse_args([])

    # Loading dot env file if provided
    if args.env:
        env_path = Path(args.env)
        if env_path.is_file():
            print(f'loading environment from {env_path.resolve()}')
            dotenv.load_dotenv(env_path, verbose=True, override=True)
        else:
            raise FileNotFoundError(f'.env file not found at {env_path} ')

    # Setting environment variables in the following priority from highest to lowest:
    # Provided commandline argument -> Dotenv environment variable -> System environment variable -> Default value

    def resolve_arg(arg_val, env_var, arg_default_val, create_directory=False):
        # Check if argument has been provided
        if not arg_val == arg_default_val:
            val = arg_val
        # Check if environment variable exists
        elif os.environ.get(env_var):
            val = os.environ[env_var]
        # Return default
        else:
            val = arg_default_val
        print(f"\t{env_var} : {val}")

        if create_directory:
            Path(val).mkdir(parents=False, exist_ok=True)

        return val

    # Write values to OS variables
    os.environ[env.NOVA_SERVER_HOST] = resolve_arg(
        args.host, env.NOVA_SERVER_HOST, default_args.host
    )
    os.environ[env.NOVA_SERVER_PORT] = resolve_arg(
        args.port, env.NOVA_SERVER_PORT, default_args.port
    )

    os.environ[env.NOVA_SERVER_CML_DIR] = resolve_arg(
        args.cml_dir, env.NOVA_SERVER_CML_DIR, default_args.cml_dir, create_directory=True
    )
    os.environ[env.NOVA_SERVER_DATA_DIR] = resolve_arg(
        args.data_dir, env.NOVA_SERVER_DATA_DIR, default_args.data_dir, create_directory=True
    )
    os.environ[env.NOVA_SERVER_CACHE_DIR] = resolve_arg(
        args.cache_dir, env.NOVA_SERVER_CACHE_DIR, default_args.cache_dir, create_directory=True
    )
    os.environ[env.NOVA_SERVER_TMP_DIR] = resolve_arg(
        args.tmp_dir, env.NOVA_SERVER_TMP_DIR, default_args.tmp_dir, create_directory=True
    )
    os.environ[env.NOVA_SERVER_LOG_DIR] = resolve_arg(
        args.log_dir, env.NOVA_SERVER_LOG_DIR, default_args.log_dir, create_directory=True
    )
    print("...done")

    host = os.environ[env.NOVA_SERVER_HOST]
    port = int(os.environ[env.NOVA_SERVER_PORT])

    serve(app, host=host, port=port)

if __name__ == '__main__':
    _run()



""" Blueprint for data extraction

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

This module defines a Flask Blueprint for extracting data.

"""


from pathlib import Path
from flask import Blueprint, request, jsonify
from nova_server.utils.thread_utils import THREADS
from nova_server.utils.job_utils import get_job_id_from_request_form
from nova_server.utils import thread_utils, job_utils
from nova_server.utils import log_utils
from nova_server.exec.execution_handler import NovaExtractHandler

extract = Blueprint("extract", __name__)


@extract.route("/extract", methods=["POST"])
def predict_thread():
    """
    Start a data extraction job.

    This route allows starting a data extraction job by providing the required parameters in the request.

    Returns:
        dict: A JSON response indicating the success of the job initiation.

    Example:
        >>> POST /extract
        >>> {"param1": "value1", "param2": "value2"}
        {"success": "true"}
    """
    if request.method == "POST":
        request_form = request.form.to_dict()
        key = get_job_id_from_request_form(request_form)
        job_utils.add_new_job(key, request_form=request_form)
        thread = extract_data(request_form)
        thread.start()
        THREADS[key] = thread
        data = {"success": "true"}
        return jsonify(data)


@thread_utils.ml_thread_wrapper
def extract_data(request_form):
    """
    Extract data in a separate thread.

    Args:
        request_form (dict): A dictionary containing the request parameters.

    Returns:
        None

    This function runs the data extraction process in a separate thread.

    """
    key = get_job_id_from_request_form(request_form)

    job_utils.update_status(key, job_utils.JobStatus.RUNNING)
    logger = log_utils.get_logger_for_job(key)
    handler = NovaExtractHandler(request_form, logger=logger)

    # TODO replace .env with actual path
    dotenv_path = Path(".env").resolve()

    try:
        handler.run(dotenv_path)
        job_utils.update_status(key, job_utils.JobStatus.FINISHED)
    except:
        job_utils.update_status(key, job_utils.JobStatus.ERROR)



"""Utility functions to create virtual environments and execute commands in them

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

"""
import os
import json
import platform
from pathlib import Path
from nova_server.utils import env
from importlib.machinery import SourceFileLoader


def _args_run_cmd(args: list = None, kwargs: dict = None) -> str:
    """
    Generate a command string from a list of arguments and keyword arguments.

    Args:
        args (list, optional): List of arguments.
        kwargs (dict, optional): Dictionary of keyword arguments.

    Returns:
        str: Combined command string.

    Example:
        >>> _args_run_cmd(['arg1', 'arg2'], {'--flag': 'value'})
        'arg1 arg2 --flag value'
    """
    tmp = []
    if args is not None:
        tmp += args
    if kwargs is not None:
        tmp += [f"{k} {json.dumps(v)}" for k, v in kwargs.items()]
    args_run_cmd = " ".join(tmp)

    return args_run_cmd


def _src_activate_cmd(env_path: Path):
    """
    Generate the source command to activate a virtual environmend. Generated output is platform dependend.

    Args:
        env_path (Path): Path to the virtual environment.

    Returns:
        str: Source activation command.

    Example:
        >>> _src_activate_cmd(Path('/path/to/venv'))
        'source /path/to/venv/bin/activate'
    """
    if platform.system() == "Windows":
        return f"{env_path/'Scripts'/'activate'}"
    else:
        return f"source {env_path/'bin'/'activate'}"


def get_module_run_cmd(
    env_path: Path, module: str, args: list = None, kwargs: dict = None
):
    """
    Generate a command to run a Python module within a virtual environment.

    Args:
        env_path (Path): Path to the virtual environment.
        module (str): Python module name.
        args (list, optional): List of arguments to pass to the module.
        kwargs (dict, optional): Dictionary of keyword arguments to pass to the module.

    Returns:
        str: Run command.

    Example:
        >>> get_module_run_cmd(Path('/path/to/venv'), 'mymodule', ['arg1', 'arg2'], {'--flag': 'value'})
        'source /path/to/venv/bin/activate && python -m mymodule arg1 arg2 --flag value'
    """
    return f"{_src_activate_cmd(env_path)} && python -m {module} {_args_run_cmd(args, kwargs)}"


def get_python_script_run_cmd(
    env_path: Path, script: Path, args: list = None, kwargs: dict = None
):
    """
    Generate a command to run a Python script within a virtual environment.

    Args:
        env_path (Path): Path to the virtual environment.
        script (Path): Path to the Python script.
        args (list, optional): List of arguments to pass to the script.
        kwargs (dict, optional): Dictionary of keyword arguments to pass to the script.

    Returns:
        str: Run command.

    Example:
        >>> get_python_script_run_cmd(Path('/path/to/venv'), Path('/path/to/script.py'), ['arg1', 'arg2'], {'--flag': 'value'})
        'source /path/to/venv/bin/activate && python /path/to/script.py arg1 arg2 --flag value'
    """
    return f"{_src_activate_cmd(env_path)} && python {script.resolve()} {_args_run_cmd(args, kwargs)}"

def get_shell_script_run_cmd(
        env_path: Path, script: str, args: list = None, kwargs: dict = None
):
    """
    Generate a command to run a console script within a virtual environment. The path to the script musst be set in the path environment variable of the console session.

    Args:
        env_path (Path): Path to the virtual environment.
        script (Path): Path to the Python script.
        args (list, optional): List of arguments to pass to the script.
        kwargs (dict, optional): Dictionary of keyword arguments to pass to the script.

    Returns:
        str: Run command.

    Example:
        >>> get_python_script_run_cmd(Path('/path/to/venv'), Path('/path/to/script.py'), ['arg1', 'arg2'], {'--flag': 'value'})
        'source /path/to/venv/bin/activate && python /path/to/script.py arg1 arg2 --flag value'
    """
    return f"{_src_activate_cmd(env_path)} && {script} {_args_run_cmd(args, kwargs)}"


def _venv_name_from_mod(module_dir: Path) -> str:
    """
    Generate a virtual environment name from a nova server module directory.
    The name equals the root folder name of the model. If a version.py file is present at the top level od the module
    directory the returned name will be equal to <root_folder_name>@<major_version>.<minor_version>.<patch_version>
    Args:
        module_dir (Path): Path to the module directory.

    Returns:
        str: Virtual environment name.
    """
    venv_name = module_dir.name
    version_file = module_dir / "version.py"
    if version_file.is_file():
        v = SourceFileLoader("version", str(version_file.resolve())).load_module()
        venv_name += f"@{v.__version__}"

    return venv_name


def venv_dir_from_mod(module_dir: Path) -> Path:
    """
    Returns the path to a virtual environment directory matchin a provided module directory.

    Args:
        module_dir (Path): Path to the module directory.

    Returns:
        Path: Virtual environment directory.

    Raises:
        ValueError: If the NOVA_CACHE_DIR environment variable is not set.

    Example:
        >>> venv_dir_from_mod(Path('/path/to/my_module'))
        Path('/path/to/venvs/my_module')
    """
    parent_dir = os.getenv(env.NOVA_SERVER_CACHE_DIR)

    if parent_dir is None:
        raise ValueError("NOVA_CACHE_DIR environment variable has not been set")

    parent_dir = Path(parent_dir) / "venvs"
    venv_dir = parent_dir / _venv_name_from_mod(module_dir)
    return venv_dir



"""Custom dependency handling for NOVA-Server modules

Authors:
    Dominik Schiller <dominik.schiller@uni-a.de>,
    Tobias Hallmen <tobias.hallmen@informatik.uni-augsburg.de>
Date:
    13.09.2023
"""

import sys
from pathlib import Path
import subprocess
import site

def assert_or_install_dependencies(packages, trainer_name):
    site_package_path = (
        Path(site.getsitepackages()[0])
        / ".."
        / "nova-server-site-packages"
        / trainer_name
    ).resolve()
    site_package_path.mkdir(parents=True, exist_ok=True)

    #TODO: handle pip install *.zip
    for i, pkg in enumerate(packages):
        params = []
        # split on space while also removing double spaces
        pk = [x for x in pkg.split(" ") if x]
        if len(pk) > 1:
            params.extend(pk[1:])
        # maybe add all VCS https://pip.pypa.io/en/stable/topics/vcs-support/
        if "git+" in pk[0]:
            if "#egg=" in pk[0]:
                name = pk[0].split("#egg=")[-1]
            else:
                name = pk[0].split("/")[-1].split(".git")[0]
        else:
            # maybe support all specifiers https://peps.python.org/pep-0440/#version-specifiers
            name = pk[0].split('==')[0]
            # setup-version-specifier
            if '[' in name:
                name = name[:name.find('[')]

        # support systems without nvidia/cuda installed
        try:
            subprocess.check_output("nvidia-smi")
            cuda_available = True
        except:
            cuda_available = False

        # torch/torchvision/torchaudio
        # remove cuda specific torch wheel params
        if name.startswith("torch") and "+cu" in name and not cuda_available:
            pk[0] = pk[0][: pk[0].find("+cu")]
            for p in pk:
                if "index-url" in p or "download.pytorch.org/whl" in p:
                    params.remove(p)

        # onnxruntime
        if name == 'onnxruntime':
            if cuda_available:
                pk[0] = pk[0].replace('onnxruntime', 'onnxruntime-gpu')
            elif sys.platform == 'darwin':
                pk[0] = pk[0].replace('onnxruntime', 'onnxruntime-silicon')

        params.append("--target={}".format(site_package_path))

        # appears to be pythonic way (snake case)
        if Path(
            "{}/{}".format(
                site_package_path, adjusted_name := str(name).replace("-", "_")
            )
        ).exists() or any(
            adjusted_name in x
            for x in [
                x.name
                for x in Path(f"{site_package_path}").glob(
                    f"{adjusted_name}-*.dist-info"
                )
            ]
        ):
            print(
                f"Skip installation of {site_package_path}/{name} - package already installed"
            )
        # some don't follow it, e.g. pyannote (. instead of _)
        elif Path(
            "{}/{}".format(
                site_package_path, adjusted_name := str(name).replace("-", ".")
            )
        ).exists() or any(
            adjusted_name in x
            for x in [
                x.name
                for x in Path(f"{site_package_path}").glob(
                    f"{adjusted_name}-*.dist-info"
                )
            ]
        ):
            print(
                f"Skip installation of {site_package_path}/{name} - package already installed"
            )
        else:
            install_package(pk[0], params)

    sys.path.insert(0, str(site_package_path.resolve()))


def install_package(pkg, params):
    call = [sys.executable, "-m", "pip", "install", pkg, *params]
    print(*call)
    return subprocess.check_call(call)



""" Blueprint for canceling a job

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

This module defines a Flask Blueprint for canceling a job.

"""

from flask import Blueprint, request, jsonify
from nova_server.utils import job_utils
from nova_server.utils.job_utils import get_job_id_from_request_form
from nova_server.utils.thread_utils import THREADS
from nova_server.utils.log_utils import LOGS


cancel = Blueprint("cancel", __name__)


@cancel.route("/cancel", methods=["POST"])
def complete_thread():
    """
    Cancel a running job.

    This route allows canceling a running job by providing the job's unique identifier in the request.

    Returns:
        dict: A JSON response indicating whether the cancelation was successful.

    Example:
        >>> POST /cancel
        >>> {"job_id": "12345"}
        {"success": "true"}
    """
    if request.method == "POST":
        request_form = request.form.to_dict()
        key = get_job_id_from_request_form(request_form)

        if key in THREADS:
            thread = THREADS[key]
            thread.raise_exception()
            job_utils.update_status(key, job_utils.JobStatus.WAITING)
            if key in LOGS:
                log = LOGS[key]
                log.info("Action successfully canceled.")
            return jsonify({"success": "true"})
        else:
            if key in LOGS:
                log = LOGS[key]
                log.info("Cancel was not successful.")
            return jsonify({"success": "false"})



""" Blueprint for data prediction

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

This module defines a Flask Blueprint for predicting data.

"""


from pathlib import Path
from flask import Blueprint, request, jsonify
from nova_server.utils.thread_utils import THREADS
from nova_server.utils.job_utils import get_job_id_from_request_form
from nova_server.utils import thread_utils, job_utils
from nova_server.utils import log_utils
from nova_server.exec.execution_handler import NovaPredictHandler

predict = Blueprint("predict", __name__)


@predict.route("/predict", methods=["POST"])
def predict_thread():
    if request.method == "POST":
        request_form = request.form.to_dict()
        key = get_job_id_from_request_form(request_form)
        job_utils.add_new_job(key, request_form=request_form)
        thread = predict_data(request_form)
        thread.start()
        THREADS[key] = thread
        data = {"success": "true"}
        return jsonify(data)


@thread_utils.ml_thread_wrapper
def predict_data(request_form):
    job_id = get_job_id_from_request_form(request_form)

    job_utils.update_status(job_id, job_utils.JobStatus.RUNNING)
    logger = log_utils.get_logger_for_job(job_id)
    handler = NovaPredictHandler(request_form, logger=logger)

    try:
        handler.run()
        job_utils.update_status(job_id, job_utils.JobStatus.FINISHED)
    except Exception as e:
        logger.critical(f"Job failed with exception {str(e)}")
        job_utils.update_status(job_id, job_utils.JobStatus.ERROR)



from flask import Blueprint, render_template
from nova_server.utils import job_utils

ui = Blueprint("ui", __name__)


@ui.route("/")
def index():
    jobs = job_utils.get_all_jobs()
    return render_template("ajax_template.html", title="Current Jobs", jobs=jobs)


@ui.route("/data")
def data():
    return {"data": job_utils.get_all_jobs()}



"""Utility modules for NOVA-Server Logs

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    13.09.2023

"""

import logging
import threading
from pathlib import Path
import os
from nova_server.utils import job_utils, env

LOGS = {}


def get_log_conform_request(request_form):
    log_conform_request = dict(request_form)
    log_conform_request["password"] = "---"
    return log_conform_request


def get_log_path_for_thread(job_id):
    log_dir = os.environ[env.NOVA_SERVER_LOG_DIR]
    return Path(log_dir) / (job_id + ".log")


def init_logger(logger, job_id):
    print("Init logger" + str(threading.current_thread().name))
    try:
        log_path = get_log_path_for_thread(job_id)
        job_utils.set_log_path(job_id, log_path)
        handler = logging.FileHandler(log_path, "w")
        handler.setFormatter(
            logging.Formatter(
                fmt="%(asctime)s %(levelname)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
            )
        )
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        LOGS[job_id] = logger
        return logger
    except Exception as e:
        print(
            "Logger for {} could not be initialized.".format(
                str(threading.current_thread().name)
            )
        )
        raise e


def get_logger_for_job(job_id):
    logger = logging.getLogger(job_id)
    if not logger.handlers:
        logger = init_logger(logger, job_id)
    return logger


def remove_log_from_dict(job_id):
    LOGS.pop(job_id)
